#include <SPI.h>
#include <LiquidCrystal.h>

/**************Definitionen*********************/

#define Ventil1 A5
#define Ventil2 12
#define Ventil3 2
#define VentilAus 3
#define PIN_CS 10
#define rs 4
#define en 5
#define d4 6 
#define d5 7
#define d6 8
#define d7 9




LiquidCrystal lcd(rs, en, d4, d5, d6, d7);


const int GAIN_1 = 0x1;                   
const int GAIN_2 = 0x0;                   

//Variablen für PID Unterprogramm
unsigned long lastTime=0;
double Input, Output, Setpoint;
double errSum, lastErr=0;
double kp, ki, kd;
int SampleTime = 1000; //1 sec
int Zulauf = 0, Ablauf = 0, Sollwert; 

//Menü-Variablen
int menusize = 6;
String menu[] = {
  "Hauptmen\365",           //0
  "Pegel anzeigen",         //1
  "Ventilzust\341nde",      //2
  "Ventilsteuerung",        //3
  "Zulauf",                 //4
  "Ablauf",                 //5
  "Regelung",               //6
  "Sollwert"                /*7*/};
byte currentPos=1, lastPos=2, ParentPos=0, lastParent=0, lastMux0=0, lastMux1=0, lastMux3=0, Taster=0, VentilPos = 2;
bool Last5=1, Last6=1, Last7=1, MenuSub=0, Regelung=1;




void setup()
{
  Serial.begin(9600);
  Serial.print("EGS-Pruefung _3190XX");
  pinMode(PIN_CS, OUTPUT);              
  pinMode(Ventil1, OUTPUT);             
  pinMode(Ventil2, OUTPUT);             
  pinMode(Ventil3, OUTPUT);             
  pinMode(VentilAus, OUTPUT);           
  DDRC = 0xEF;                          
  
  
  SPI.begin();
  SPI.setClockDivider(SPI_CLOCK_DIV2);

  //Bei Start alle Ventile schließen
  digitalWrite(Ventil1, LOW);
  digitalWrite(Ventil2, LOW);
  digitalWrite(Ventil3, LOW);
  digitalWrite(VentilAus, LOW);
  

  setOutput(0, GAIN_2, 1, 0);              
  Sollwert = 50;                           
  Setpoint = map(Sollwert,0,100,0,4096);   
  
  //LCD Start mit Begrüßung
  lcd.begin(16, 4);
  lcd.print("PAL Pr\365fung");
  lcd.setCursor(0, 1);
  lcd.print("G\341rtnerei");
  delay(900);
  lcd.clear();
 
  //Werte für PID Ki, Kp und Kd
  SetTunings(16, 0.0002, 0);

}

void setOutput(bool ign, bool gain, bool SHDN, unsigned int val)
{
  byte lowByte = val & 0xff;
  byte highByte = ((val >> 8) & 0xff) | ign << 7 | gain << 5 | SHDN << 4;

  digitalWrite(PIN_CS,LOW);
  SPI.transfer(highByte);
  SPI.transfer(lowByte);
  digitalWrite(PIN_CS,HIGH);
}

int Mux(int kanal)
{
  switch (kanal)
  {
    case 0:
      digitalWrite(A3, LOW);
      digitalWrite(A2, LOW);
      digitalWrite(A1, LOW);
      break;

    case 1:
      digitalWrite(A3, HIGH);
      digitalWrite(A2, LOW);
      digitalWrite(A1, LOW);
      break;

    case 2:
      digitalWrite(A3, LOW);
      digitalWrite(A2, HIGH);
      digitalWrite(A1, LOW);
      break;

    case 3:
      digitalWrite(A3, HIGH);
      digitalWrite(A2, HIGH);
      digitalWrite(A1, LOW);
      break;
      
    case 4:
      digitalWrite(A3, LOW);
      digitalWrite(A2, LOW);
      digitalWrite(A1, HIGH);
      break;
      
    case 5:
      digitalWrite(A3, HIGH);
      digitalWrite(A2, LOW);
      digitalWrite(A1, HIGH);
      break;
      
    case 6:
      digitalWrite(A3, LOW);
      digitalWrite(A2, HIGH);
      digitalWrite(A1, HIGH);
      break;

    case 7:
      digitalWrite(A3, HIGH);
      digitalWrite(A2, HIGH);
      digitalWrite(A1, HIGH);
      break;

    default:
      break;
  }
    delay(5);

  if((kanal == 5) || (kanal == 6) || (kanal == 7))
  {
    return digitalRead(A4);
  }
  else
  {
    return map(analogRead(A4), 0, 1024, 0, 5000);
  }
}

void Compute()
{
   unsigned long now = millis();
   int timeChange = (now - lastTime);
   if(timeChange>=SampleTime)
   {
    double error = Setpoint - Input;
    errSum += (error * timeChange);
    double dErr = (error - lastErr) / timeChange;
  
    Output = kp * error + ki * errSum;// + kd * dErr;
    if(Output>=4095)Output=4095;
    if(Output<=0)Output=0;
  
    lastErr = error;
    lastTime = now;
   }
}

void SetTunings(double Kp, double Ki, double Kd)
{
   double SampleTimeInSec = ((double)SampleTime)/1000;
   kp = Kp;
   ki = Ki * SampleTimeInSec;
   kd = Kd / SampleTimeInSec;
}

void SetSampleTime(int NewSampleTime)
{
   if (NewSampleTime > 0)
   {
      double ratio  = (double)NewSampleTime / (double)SampleTime;
      ki *= ratio;
      kd /= ratio;
      SampleTime = (unsigned long)NewSampleTime;
   }
}

void   readButtons(void)
{
  if(Mux(6) == LOW)
  {
    if(Last6 == HIGH)
    {
      delay(10);
      if(Mux(6) == LOW)
      {
        Last6 = LOW;
        Taster = Taster | 0b00000100;
      }
    }
  }
  else{Last6=HIGH;}

  if(Mux(5) == LOW)
  {
    if(Last5 == HIGH)
    {
      delay(10);
      if(Mux(5) == LOW)
      {
        Last5 = LOW;
        Taster = Taster | 0b00000001;
      }
    }
  }
  else{Last5=HIGH;}

   if(Mux(7) == LOW)
  {
    if(Last7 == HIGH)
    {
      delay(10);
      if(Mux(7) == LOW)
      {
        Last7 = LOW;
        Taster = Taster | 0b00000010;
      }
    }
  }
  else{Last7=HIGH;}

}

void  LCD_Menue(void)
{
  if(MenuSub == 1)
  {
    lastParent = ParentPos;
    ParentPos = currentPos;
    lastPos += 1;
    MenuSub = 0;
  }

  if((Taster != 0) || (lastPos != currentPos)||((ParentPos == 1)&&((lastMux1 != map(Mux(1),0,5000,0,8)||(lastMux0 != map(Mux(0),0,5000,0,8))||(lastMux3 != map(Mux(3),0,5000,0,8))))))  
  {

    switch(ParentPos)
    {
      case 0: //Hauptmenü
        //Tasterauswertung
        if(Taster == 1)
        {
          if(currentPos == 1)
          {
             currentPos = (sizeof(menu) / sizeof(menu[0]))-1;
          }
          else{currentPos--;}
        }
        if(Taster == 4)
        {
          currentPos++;
          if(currentPos > (sizeof(menu) / sizeof(menu[0]))-1)
          {
            currentPos = 1;
          }
        }
        if(Taster == 2)
        {
          MenuSub = 1;
        }
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        
        lcd.setCursor(0, 2);
        lcd.print("\176");
        lcd.print(menu[currentPos]);
        lcd.setCursor(0, 3);
        lcd.print("\74");
        lcd.setCursor(15, 3);
        lcd.print("\76");
        break;
        
      case 1: //Pegelanzeigen
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }
        
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        lcd.setCursor(0, 1);
        lcd.print("Tank 1: ");
        for( int i=0; i<map(Mux(1),0,5000,0,8);i++)
        {
          lcd.print("\377");
        }
        lastMux1 = map(Mux(1),0,5000,0,8);
        lcd.setCursor(0, 2);
        lcd.print("Tank 2: ");
        for( int i=0; i<map(Mux(0),0,5000,0,8);i++)
        {
          lcd.print("\377");
        }
        lastMux0 = map(Mux(0),0,5000,0,8);
        lcd.setCursor(0, 3);
        lcd.print("Tank 3: ");
        for( int i=0; i<map(Mux(3),0,5000,0,8);i++)
        {
          lcd.print("\377");
        }
        lastMux3 = map(Mux(3),0,5000,0,8);
        break;
        
      case 2: //Ventilzustände
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        lcd.setCursor(0, 1);
        lcd.print("Ventil 1:  ");
        if(digitalRead(Ventil1) == LOW)
        {
          lcd.print("ZU ");
        }
        else
        {
          lcd.print("OFFEN ");
        }
        lcd.setCursor(0, 2);
        lcd.print("Ventil 2:  ");
        if(digitalRead(Ventil2) == LOW)
        {
          lcd.print("ZU ");
        }
        else
        {
          lcd.print("OFFEN ");
        }
        lcd.setCursor(0, 3);
        lcd.print("Ventil 3:  ");
        if(digitalRead(Ventil3) == LOW)
        {
          lcd.print("ZU ");
        }
        else
        {
          lcd.print("OFFEN ");
        }
        break;
        
        case 3: 
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }

        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        if(Regelung == HIGH)
        {
          lcd.setCursor(0, 1);
          lcd.print("Ventile nur ohne");
          lcd.setCursor(0, 2);
          lcd.print("Regelung ein-");
          lcd.setCursor(0, 3);
          lcd.print("stellbar!");
        }
        else
        { //Tasterauswertung
          if(Taster == 4)
          {
            switch(VentilPos)
            {
//              case 1:
//                digitalWrite(Ventil1,~(digitalRead(Ventil1)));
//              break;
              case 2:
                digitalWrite(Ventil2,(digitalRead(Ventil2)^1));
              break;
              case 3:
                digitalWrite(Ventil3,(digitalRead(Ventil3)^1));
              break;
              default:
              break;
            }
          }
          if(Taster == 2)
          {
            VentilPos++;
            if(VentilPos > 3)
            {
              VentilPos = 2;
            }
          }
  
          switch(VentilPos)
          {
            case 2:
              lcd.setCursor(0, 1);
              lcd.print("Ventil 2:  ");
              if(digitalRead(Ventil2) == LOW)
              {
                lcd.print("ZU ");
              }
              else
              {
                lcd.print("OFFEN ");
              }
  
            break;
  
            case 3:
              lcd.setCursor(0, 1);
              lcd.print("Ventil 3:  ");
              if(digitalRead(Ventil3) == LOW)
              {
                lcd.print("ZU ");
              }
              else
              {
                lcd.print("OFFEN ");
              }
  
            break;
  
            default:
              break;
          }
          lcd.setCursor(0, 3);
          lcd.print("Ventil ");
          lcd.setCursor(7, 3);
          lcd.print("\76");
          lcd.setCursor(10, 3);
          switch(VentilPos)
          {
            case 2:
              if(digitalRead(Ventil2) == LOW)
              {
                lcd.print("\357ffnen");
              }
              else lcd.print("  Zu");
            break;
            case 3:
              if(digitalRead(Ventil3) == LOW)
              {
                lcd.print("\357ffnen");
              }
              else lcd.print("  Zu");
            break;
            default:
            break;
          }
        }
        break;

        case 4: //Zulauf
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }
        if(Taster == 4)
        {
          if(Zulauf < 100)
          {
            Zulauf += 10;
            if(Zulauf > 100) Zulauf = 100;
          }
        }
        if(Taster == 2)
        {
          if(Zulauf > 0)
          {
            Zulauf -= 10;
            if(Zulauf < 0) Zulauf = 0;
          }
        }
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        if(Regelung == HIGH)
        {
          lcd.setCursor(0, 1);
          lcd.print("Zulauf nur ohne ");
          lcd.setCursor(0, 2);
          lcd.print("Regelung ein-");
          lcd.setCursor(0, 3);
          lcd.print("stellbar!");
        }
        else
        {
          lcd.setCursor(0, 1);
          lcd.print("Zulauf: ");
          lcd.print(Zulauf);
          lcd.setCursor(11, 1);
          lcd.print("%");
          lcd.setCursor(7,3);
          lcd.print("-");
          lcd.setCursor(13,3);
          lcd.print("+");
        }
        break;


        case 5: //Ablauf
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }        
        if(Taster == 4)
        {
          if(Ablauf < 100)
          {
            Ablauf += 5;
            if(Ablauf > 100) Ablauf = 100;
          }
        }
        if(Taster == 2)
        {
          if(Ablauf > 0)
          {
            Ablauf -= 5;
            if(Ablauf < 0) Ablauf = 0;
          }
        }
        
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");

        lcd.setCursor(0, 1);
        lcd.print("Ventil Aus:");
        if(Ablauf > 0)
        {
          lcd.print("OFFEN ");
        }
        else
        {
          lcd.print("ZU ");
        }
  
        lcd.setCursor(0, 2);
        lcd.print("Ablauf: ");
        lcd.print(Ablauf);
        lcd.setCursor(11, 2);
        lcd.print("%");
        lcd.setCursor(7,3);
        lcd.print("-");
        lcd.setCursor(13,3);
        lcd.print("+");
        
        break;


        case 6: //Regelung
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }
        if(Taster == 2)
        {
          Regelung = HIGH;
        }
        if(Taster == 4)
        {
          Regelung = LOW;
        }
        
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        lcd.setCursor(0, 1);
        lcd.print("Regelung: ");
        if(Regelung == HIGH)
        {
          lcd.print("AN");
        }
        else
        {
          lcd.print("AUS");
        }
        lcd.setCursor(0, 2);
        lcd.print("Sollwert: ");
        lcd.print(Sollwert);
        lcd.setCursor(13, 2);
        lcd.print("%");
        lcd.setCursor(7, 3);
        lcd.print("AN");
        lcd.setCursor(12, 3);
        lcd.print("AUS");
        break;

        case 7: //Sollwert
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }
        if(Taster == 2)
        {
          //Setpoint = Setpoint - 41;
          Sollwert = Sollwert - 5;
          Setpoint = map(Sollwert,0,100,0,4096);
          if(Setpoint < 0)
            Sollwert = 0;
            Setpoint = 0;
        }
        if(Taster == 4)
        {
          //Setpoint = Setpoint + 41;
          Sollwert = Sollwert + 5;
          Setpoint = map(Sollwert,0,100,0,4096);
          if(Setpoint > 4096)
            Sollwert = 100;
            Setpoint = 4096;
        }
        
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        lcd.setCursor(0, 1);
        lcd.print("Regelung: ");
        if(Regelung == HIGH)
        {
          lcd.print("AN");
        }
        else
        {
          lcd.print("AUS");
        }
        lcd.setCursor(0, 2);
        lcd.print("Sollwert: ");
        lcd.print(Sollwert);
        lcd.setCursor(13, 2);
        lcd.print("%");
        lcd.setCursor(7, 3);
        lcd.print("-");
        lcd.setCursor(13, 3);
        lcd.print("+");
        break;
        

       default:
        if(Taster == 1)
        {
          ParentPos = lastParent;
          lastPos++;
          
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print(menu[ParentPos]);
          lcd.setCursor(0, 2);
          lcd.print("\176");
          lcd.print(menu[currentPos]);
          lcd.setCursor(0, 3);
          lcd.print("\74");
          lcd.setCursor(15, 3);
          lcd.print("\76");
          break;
        }
        //LCD Ansteuerung
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(menu[ParentPos]);
        lcd.print(":");
        break;
    }
    lastPos = currentPos;
  }
  Taster = 0;
}

void loop()
{
  if(Regelung == HIGH)
  {
    Input=Mux(1);
    Compute();
    if(Output>0)  
    {
      digitalWrite(Ventil1,HIGH);
      digitalWrite(Ventil2,LOW);
    }
    setOutput(0, GAIN_2, 1, Output);
    if(Mux(0) >= Mux(1))
    {
      digitalWrite(Ventil2,HIGH);
    }
    else
    {
      digitalWrite(Ventil2,LOW);
    }
    if(Mux(3) >= Mux(0))
    {
      digitalWrite(Ventil3, HIGH);
    }
    else
    {
      digitalWrite(Ventil3,LOW);
    }
  }
  else
  {
    if(Zulauf >= 1)
    {
      if(Mux(1) - Mux(2) > 0)
      {
        int x = (Zulauf+map(Mux(1),0,1024,0,100));          
        setOutput(0, GAIN_2, 1, map(x,0,(x/100+x%100),0,4095));   
      }
      else
      {
        setOutput(0, GAIN_2, 1, map(Zulauf,0,100,0,4095));
      }
      digitalWrite(Ventil1,HIGH);
    }
    else
    {
      setOutput(0, GAIN_2, 1, 0);
      digitalWrite(Ventil1,LOW);
    }
  }
  analogWrite(VentilAus, map(Ablauf,0,100,0,50));
  readButtons();
  LCD_Menue();
}
